"""
NoORM adapter for synchronous access to MySQL/MariaDB via PyMySQL.

Decorators:
- `sql_fetch_all(row_type: type, sql: str)` to fetch records as a list
- `sql_one_or_none(res_type: type, sql: str)` to fetch one record
- `sql_scalar_or_none(res_type: type, sql: str)` to fetch a scalar
- `sql_fetch_scalars(res_type: type, sql: str)` to fetch a list of scalars
- `sql_execute(sql: str)` to execute a statement
- `sql_iterate(row_type: type, sql: str)` and
  `sql_iterate_scalars(row_type: type, sql: str)` to make a query and iterate
  through results, objects or scalars respectively.

Decorator parameters are:
- (except `sql_execute`) Expected result type. For `sql_fetch_all` and `sql_one_or_none`
  is usually a dataclass or namedtuple; for `sql_scalar_or_none` and `sql_fetch_scalars`
  is usually `int`, `str`, `bool`, `datetime`, or whatever can be produced by scalar
  query.
- SQL statement as string. If None, must be provided from the decorated function.
  Use `%s` or `%(name)s` as parameters placeholders.

After decoration the decorated function receives an open PyMySQL DB connection as a
first positional argument.

Decorated function should return:
- `None` if no changes to original sql statement is needed,
  and no params to be applied
- Use `params`, `query_only`, or `query_and_params` functions to return accordingly
  parameters values, SQL statement, or both together.
  - `params(*args, **kwargs)` to pass positional arguments for `%s`-style placeholders
    or keyword arguments for named query parameter placeholders. SQL statement is
    original from decorator parameter.
  - `query_only(sql: str)` - only SQL query and no query parameters.
  - `query_and_params(sql: str, *args, **kwargs)` to provide both - a query and its
    parameters.

Examples:
```
from collections import namedtuple
import pymysql

import noorm.pymysql as nm

# Fetch all, no parameters
@nm.sql_fetch_all(
    namedtuple("DbAllUsersResult", "id, username"),
    "select id, username from users order by id;",
)
def get_all_users():
    pass  # no parameters for this query

# Fetch one by id
@nm.sql_one_or_none(
    namedtuple("DbOneUserResult", "username, email"),
    "select username, email from users where id = %s;"
)
def get_one_user(id_: int):
    return nm.params(id_)

# Fetch scalar
@nm.sql_scalar_or_none(int, "select count(*) from users where username ilike %(search)s;")
def get_num_users(search: str):
    return nm.params(search=f"%{search}%")

# Insert a new record
@nm.sql_execute("insert into users(username, email) values(%s, %s)")
def ins_user(username: str | None = None, email: str | None = None):
    return nm.params(username, email)

# Usage:
with pymysql.connect(
    host="<<DB_HOST>>", user="<<USER>>", password="<<PWD>>", database="<<DB_NAME>>"
) as conn:
    for user in get_all_users(conn):
        print(user)
    print(f"User with id=1: {get_one_user(conn, 1)}")
    print(f"Search results by 'John': {get_num_users(conn, 'John')}")

    ins_user(conn, "Jane", "jane@example.com")
    conn.commit()  # Do not forget to commit the data manipulation!!!
```
"""  # noqa: E501

from ._pymysql import (
    sql_fetch_all,
    sql_iterate,
    sql_one_or_none,
    sql_scalar_or_none,
    sql_fetch_scalars,
    sql_iterate_scalars,
    sql_execute,
)
from noorm._db_api_2 import params, query_and_params, query_only
from noorm._common import CancelExecException

__all__ = [
    "sql_fetch_all",
    "sql_iterate",
    "sql_one_or_none",
    "sql_scalar_or_none",
    "sql_fetch_scalars",
    "sql_iterate_scalars",
    "sql_execute",
    "params",
    "query_and_params",
    "query_only",
    "CancelExecException",
]



"""
NoORM (Not only ORM)

Typical ways to import:
- `import noorm.sqlite3 as nm` - for SQLite
- `import noorm.aiosqlite as nm` - for asynchronous SQLite via aiosqlite
- `import noorm.psycopg2 as nm` - for synchronous Postgres via psycopg2
- `import noorm.asyncpg as nm` - for asynchronous Postgres via asyncpg
- `import noorm.pymysql as nm` - for synchronous MySQL/MariaDB via PyMySQL
- `import noorm.aiomysql as nm` - for asynchronous MySQL/MariaDB via aiomysql
- `import noorm.sqlalchemy_sync as nm` - for synchronous SqlAlchemy
- `import noorm.sqlalchemy_async as nm` - for asynchronous SqlAlchemy
"""



from typing import Type, Callable, AsyncGenerator, ParamSpec, TypeVar, Concatenate
from typing import Coroutine, Any
from typing import overload
import aiosqlite

from .._common import WrapperBase
from .._sqlite_common import (
    make_decoder as _make_decoder,
    make_scalar_decoder as _make_scalar_decoder,
    sqlite_sql_n_params,
)
from .._db_api_2 import PrepareFuncResult
from ..registry import MetricsCollector

F_Spec = ParamSpec("F_Spec")
TR = TypeVar("TR")
ConnectionOrCursor = aiosqlite.Connection | aiosqlite.Cursor


def sql_fetch_all(row_type: Type[TR], sql: str | None = None):
    """
    Use this decorator to make a "fetch all" SQL statement executor out of
    the function that prepares parameters for the query.

    :param row_type: type of expected result. Usually some dataclass or named tuple
    :param sql: SQL statement to execute. If None, the SQL statement must be provided
    by decorated function.

    IMPORTANT: decorated function must not be async, but after decoration it
    becomes async.

    More info in the noorm.aiosqlite docstring.
    """

    def decorator(
        func: Callable[F_Spec, PrepareFuncResult | None]
    ) -> Callable[
        Concatenate[ConnectionOrCursor, F_Spec], Coroutine[Any, Any, list[TR]]
    ]:
        decoder = _make_decoder(row_type)

        class wrapper(WrapperBase):
            async def __call__(
                self,
                conn: ConnectionOrCursor,
                *args: F_Spec.args,
                **kwargs: F_Spec.kwargs,
            ) -> list[TR]:
                with MetricsCollector(self._func) as mc:
                    if sql_and_params := sqlite_sql_n_params(
                        self._func, args, kwargs, sql
                    ):
                        q_res = await conn.execute(*sql_and_params)
                        col_names = tuple(el[0] for el in q_res.description)
                        res: list[TR] = [
                            row_type(**decoder({n: v for n, v in zip(col_names, r)}))
                            async for r in q_res
                        ]
                        mc.tuples = len(res)
                        return res
                    return []

        return wrapper(func)

    return decorator


def sql_iterate(row_type: Type[TR], sql: str | None = None):
    """
    Use this decorator to make a query and iterate through results. Be careful with
    this feature and, if possible, use `sql_fetch_all` instead, because
    `sql_fetch_all` gives you less possibilites to shoot your leg.

    :param row_type: type of expected result. Usually some dataclass or named tuple
    :param sql: SQL statement to execute. If None, the SQL statement must be provided
    by decorated function.

    IMPORTANT: decorated function must not be async, but after decoration it
    becomes async.

    More info in the noorm.aiosqlite docstring.
    """

    def decorator(
        func: Callable[F_Spec, PrepareFuncResult | None]
    ) -> Callable[Concatenate[aiosqlite.Connection, F_Spec], AsyncGenerator[TR, None]]:
        decoder = _make_decoder(row_type)

        class wrapper(WrapperBase):
            async def __call__(
                self,
                conn: aiosqlite.Connection,
                *args: F_Spec.args,
                **kwargs: F_Spec.kwargs,
            ) -> AsyncGenerator[TR, None]:
                with MetricsCollector(self._func) as mc:
                    if sql_and_params := sqlite_sql_n_params(
                        self._func, args, kwargs, sql
                    ):
                        cur = await conn.cursor()
                        q_res = await cur.execute(*sql_and_params)
                        col_names = tuple(el[0] for el in q_res.description)
                        is_first_row = True
                        async for r in q_res:
                            if is_first_row:
                                mc.finish(None)
                                is_first_row = False
                            yield row_type(
                                **decoder({n: v for n, v in zip(col_names, r)})
                            )

        return wrapper(func)

    return decorator


def sql_one_or_none(row_type: Type[TR], sql: str | None = None):
    """
    Use this decorator to make a "one or none" SQL statement executor out of
    the function that prepares parameters for the query.

    :param row_type: type of expected result. Usually some dataclass or named tuple
    :param sql: SQL statement to execute. If None, the SQL statement must be provided
    by decorated function.

    IMPORTANT: decorated function must not be async, but after decoration it
    becomes async.

    More info in the noorm.aiosqlite docstring.
    """

    def decorator(
        func: Callable[F_Spec, PrepareFuncResult | None]
    ) -> Callable[
        Concatenate[ConnectionOrCursor, F_Spec], Coroutine[Any, Any, TR | None]
    ]:
        decoder = _make_decoder(row_type)

        class wrapper(WrapperBase):
            async def __call__(
                self,
                conn: ConnectionOrCursor,
                *args: F_Spec.args,
                **kwargs: F_Spec.kwargs,
            ) -> TR | None:
                with MetricsCollector(self._func) as mc:
                    if sql_and_params := sqlite_sql_n_params(
                        self._func, args, kwargs, sql
                    ):
                        q_res = await conn.execute(*sql_and_params)
                        col_names = tuple(el[0] for el in q_res.description)
                        async for row in q_res:
                            mc.tuples = 1
                            return row_type(
                                **decoder({n: v for n, v in zip(col_names, row)})
                            )
                    return None

        return wrapper(func)

    return decorator


def sql_scalar_or_none(res_type: Type[TR], sql: str | None = None):
    """
    Use this decorator to make a "scalar or none" SQL statement executor out of
    the function that prepares parameters for the query.

    :param row_type: type of expected result. For scalar queries it is usually `int`,
    `str`, `bool`, `datetime`, or whatever can be produced by scalar query. Use Any to
    return a value without conversion.
    :param sql: SQL statement to execute. If None, the SQL statement must be provided
    by decorated function.

    IMPORTANT: decorated function must not be async, but after decoration it
    becomes async.

    More info in the noorm.aiosqlite docstring.
    """

    def decorator(
        func: Callable[F_Spec, PrepareFuncResult | None]
    ) -> Callable[
        Concatenate[ConnectionOrCursor, F_Spec], Coroutine[Any, Any, TR | None]
    ]:
        decoder = _make_scalar_decoder(res_type)

        class wrapper(WrapperBase):
            async def __call__(
                self,
                conn: ConnectionOrCursor,
                *args: F_Spec.args,
                **kwargs: F_Spec.kwargs,
            ) -> TR | None:
                with MetricsCollector(self._func) as mc:
                    if sql_and_params := sqlite_sql_n_params(
                        self._func, args, kwargs, sql
                    ):
                        q_res = await conn.execute(*sql_and_params)
                        async for row in q_res:
                            mc.tuples = 1
                            return decoder(row[0])
                    return None

        return wrapper(func)

    return decorator


def sql_fetch_scalars(res_type: Type[TR], sql: str | None = None):
    """
    Use this decorator to make a "fetch scalars" SQL statement executor out of
    the function that prepares parameters for the query.

    :param row_type: type of expected result. For scalar queries it is usually `int`,
    `str`, `bool`, `datetime`, or whatever can be produced in a single-column query
    result. Use Any to return a value without conversion.
    :param sql: SQL statement to execute. If None, the SQL statement must be provided
    by decorated function.

    IMPORTANT: decorated function must not be async, but after decoration it
    becomes async.

    More info in the noorm.aiosqlite docstring.
    """

    def decorator(
        func: Callable[F_Spec, PrepareFuncResult | None]
    ) -> Callable[
        Concatenate[ConnectionOrCursor, F_Spec], Coroutine[Any, Any, list[TR]]
    ]:
        decoder = _make_scalar_decoder(res_type)

        class wrapper(WrapperBase):
            async def __call__(
                self,
                conn: ConnectionOrCursor,
                *args: F_Spec.args,
                **kwargs: F_Spec.kwargs,
            ) -> list[TR]:
                with MetricsCollector(self._func) as mc:
                    if sql_and_params := sqlite_sql_n_params(
                        self._func, args, kwargs, sql
                    ):
                        q_res = await conn.execute(*sql_and_params)
                        res = [decoder(row[0]) async for row in q_res]
                        mc.tuples = len(res)
                        return res
                    return []

        return wrapper(func)

    return decorator


def sql_iterate_scalars(res_type: Type[TR], sql: str | None = None):
    """
    Use this decorator to make a query and iterate through scalar results. Be careful
    with this feature and, if possible, use `sql_fetch_scalars` instead, because
    `sql_fetch_scalars` gives you less possibilites to shoot your leg.

    :param res_type: type of expected result. For scalar queries it is usually `int`,
    `str`, `bool`, `datetime`, or whatever can be produced in a single-column query
    result. Use Any to return values without conversion.
    :param sql: SQL statement to execute. If None, the SQL statement must be provided
    by decorated function.

    More info in the noorm.aiosqlite docstring.
    """

    def decorator(
        func: Callable[F_Spec, PrepareFuncResult | None]
    ) -> Callable[Concatenate[aiosqlite.Connection, F_Spec], AsyncGenerator[TR, None]]:
        decoder = _make_scalar_decoder(res_type)

        class wrapper(WrapperBase):
            async def __call__(
                self,
                conn: aiosqlite.Connection,
                *args: F_Spec.args,
                **kwargs: F_Spec.kwargs,
            ) -> AsyncGenerator[TR, None]:
                with MetricsCollector(self._func) as mc:
                    if sql_and_params := sqlite_sql_n_params(
                        self._func, args, kwargs, sql
                    ):
                        cur = await conn.cursor()
                        q_res = await cur.execute(*sql_and_params)
                        is_first_row = True
                        async for r in q_res:
                            if is_first_row:
                                mc.finish(None)
                                is_first_row = False
                            yield decoder(r[0])

        return wrapper(func)

    return decorator


@overload
def sql_execute(
    func: Callable[F_Spec, PrepareFuncResult | None]
) -> Callable[Concatenate[ConnectionOrCursor, F_Spec], Coroutine[Any, Any, None]]:
    pass  # pragma: no cover


@overload
def sql_execute(
    sql: str | None = None,
) -> Callable[
    [Callable[F_Spec, Coroutine[Any, Any, None]]],
    Callable[Concatenate[ConnectionOrCursor, F_Spec], Coroutine[Any, Any, None]],
]:
    pass  # pragma: no cover


def sql_execute(  # type: ignore
    sql: Callable[F_Spec, PrepareFuncResult | None] | str | None = None
):
    """
    Use this decorator to execute a statement without responding a result.

    :param sql: SQL statement to execute. If None, the SQL statement must be provided
    by decorated function.

    IMPORTANT: decorated function must not be async, but after decoration it
    becomes async.

    More info in the noorm.aiosqlite docstring.
    """
    if callable(sql):
        the_func = sql
        the_sql = None
    else:
        the_func = None
        the_sql = sql

    def decorator_wrapper(sql: str | None):
        def decorator(
            func: Callable[F_Spec, PrepareFuncResult | None]
        ) -> Callable[
            Concatenate[ConnectionOrCursor, F_Spec], Coroutine[Any, Any, None]
        ]:
            class wrapper(WrapperBase):
                async def __call__(
                    self,
                    conn: ConnectionOrCursor,
                    *args: F_Spec.args,
                    **kwargs: F_Spec.kwargs,
                ) -> None:
                    with MetricsCollector(self._func):
                        if sql_and_params := sqlite_sql_n_params(
                            self._func, args, kwargs, sql
                        ):
                            await conn.execute(*sql_and_params)

            return wrapper(func)

        return decorator

    wrap_decorator = decorator_wrapper(the_sql)
    if callable(sql):
        return wrap_decorator(the_func)
    else:
        return wrap_decorator



"""
NoORM adapter for asynchronous access to Postgres via asyncpg.

Decorators:
- `sql_fetch_all(row_type: type, sql: str)` to fetch records as a list
- `sql_one_or_none(res_type: type, sql: str)` to fetch one record
- `sql_scalar_or_none(res_type: type, sql: str)` to fetch a scalar
- `sql_fetch_scalars(res_type: type, sql: str)` to fetch a list of scalars
- `sql_execute(sql: str)` to execute a statement

Decorator parameters are:
- (except `sql_execute`) Expected result type. For `sql_fetch_all` and `sql_one_or_none`
  is usually a dataclass or namedtuple; for `sql_scalar_or_none` and `sql_fetch_scalars`
  is usually `int`, `str`, `bool`, `datetime`, or whatever can be produced by scalar
  query.
- SQL statement as string. If None, must be provided from the decorated function.
  Use `$1`, `$2`, etc. as parameters placeholders.

After decoration the decorated function receives an open AsyncPg DB connection as a
first positional argument.

IMPORTANT: decorated function must not be async, but after decoration it becomes async.

Decorated function should return:
- `None` if no changes to original sql statement is needed,
  and no params to be applied
- Use `params`, `query_only`, or `query_and_params` functions to return accordingly
  parameters values, adjusted statement, or both these things together.
  - `params(*args)` to pass parameters.
  - `query_only(sql: str)` - no query parameters, but another query instead of original.
  - `query_and_params(sql: str, *args)` to provide both - a new query and parameters.

Examples:
```
from collections import namedtuple
import asyncpg

import noorm.asyncpg as nm

# Fetch all, no parameters
@nm.sql_fetch_all(
    namedtuple("DbAllUsersResult", "id, username"),
    "select id, username from users order by id;",
)
def get_all_users():  # Consider no "async"
    pass  # no parameters for this query

# Fetch one by id
@nm.sql_one_or_none(
    namedtuple("DbOneUserResult", "username, email"),
    "select username, email from users where id = $1;"
)
def get_one_user(id_: int):
    return nm.params(id_)

# Fetch scalar
@nm.sql_scalar_or_none(int, "select count(*) from users where username ilike $1;")
def get_num_users(search: str):
    return nm.params(f"%{search}%")

# Insert a new record
@nm.sql_execute("insert into users(username, email) values($1, $2)")
def ins_user(username: str | None = None, email: str | None = None):
    return nm.params(username, email)

# Usage:
conn = await asyncpg.connect('postgresql://postgres@localhost/test')
for user in await get_all_users(conn):  # Consider "await"
    print(user)
print(f"User with id=1: {await get_one_user(conn, 1)}")
print(f"Search results by 'John': {await get_num_users(conn, 'John')}")

await ins_user(conn, "Jane", "jane@example.com")
# NOTE: No need to commit because asyncpg uses auto-commit.
# Explicit commit is needed only in transactions.

await conn.close()
```
"""

from ._asyncpg import (
    sql_fetch_all,
    sql_one_or_none,
    sql_scalar_or_none,
    sql_fetch_scalars,
    sql_execute,
)
from noorm._db_api_2_args_only import params, query_and_params, query_only
from noorm._common import CancelExecException

__all__ = [
    "sql_fetch_all",
    "sql_one_or_none",
    "sql_scalar_or_none",
    "sql_fetch_scalars",
    "sql_execute",
    "params",
    "query_and_params",
    "query_only",
    "CancelExecException",
]



from ._common import CancelExecException


class PrepareFuncResult:
    def __init__(self, sql: str | None, params: tuple | None) -> None:
        self.sql = sql
        self.params = params


def params(*args):
    return PrepareFuncResult(None, args)


def query_and_params(sql: str, *args):
    return PrepareFuncResult(sql, args)


def query_only(sql: str):
    return PrepareFuncResult(sql, None)


def req_sql_n_params(
    func, f_args, f_kwargs, default_sql: str | None
) -> tuple[str, tuple] | None:
    try:
        sql_n_params: PrepareFuncResult | None = func(*f_args, **f_kwargs)
    except CancelExecException:
        return None
    if sql_n_params is not None:
        if not isinstance(sql_n_params, PrepareFuncResult):
            raise TypeError(
                f"Function {func.__name__} returned {type(sql_n_params).__name__} "
                "(expected None or result of 'params', 'query_and_params', or "
                "'query_only' functions)"
            )
        sql = sql_n_params.sql
        params = sql_n_params.params
        if sql is None:
            sql = default_sql
        if params is None:
            params = tuple()
    else:
        sql = default_sql
        params = tuple()
    if sql is None:
        raise RuntimeError(
            f"Function {func.__name__} did not return an SQL statement "
            "in its result. When SQL statement is not provided in decorator params, "
            "it should be returned by the function through the 'query_only' or "
            "'query_and_params' function."
        )
    return sql, params



from typing import Type, Callable, Coroutine, ParamSpec, TypeVar, Any, Concatenate
from typing import overload

import asyncpg

from .._common import WrapperBase
from .._db_api_2_args_only import PrepareFuncResult, req_sql_n_params
from ..registry import MetricsCollector

F_Spec = ParamSpec("F_Spec")
TR = TypeVar("TR")


def sql_fetch_all(row_type: Type[TR], sql: str | None = None):
    """
    Use this decorator to make a "fetch" SQL statement executor out of
    the function that prepares parameters for the query.

    :param row_type: type of expected result. Usually some dataclass or named tuple
    :param sql: SQL statement to execute. If None, the SQL statement must be provided
    by decorated function.

    IMPORTANT: decorated function must not be async, but after decoration it
    becomes async.

    More info in the noorm.asyncpg docstring.
    """

    def decorator(
        func: Callable[F_Spec, PrepareFuncResult | None]
    ) -> Callable[
        Concatenate[asyncpg.Connection, F_Spec], Coroutine[Any, Any, list[TR]]
    ]:
        class wrapper(WrapperBase):
            async def __call__(
                self,
                conn: asyncpg.Connection,
                *args: F_Spec.args,
                **kwargs: F_Spec.kwargs,
            ) -> list[TR]:
                with MetricsCollector(self._func) as mc:
                    if sql_and_params := req_sql_n_params(
                        self._func, args, kwargs, sql
                    ):
                        q_res = await conn.fetch(sql_and_params[0], *sql_and_params[1])
                        res: list[TR] = [
                            row_type(**{n: v for n, v in r.items()}) for r in q_res
                        ]
                        mc.tuples = len(res)
                        return res
                    return []

        return wrapper(func)

    return decorator


def sql_one_or_none(row_type: Type[TR], sql: str | None = None):
    """
    Use this decorator to make an "fetchrow" SQL statement executor out of
    the function that prepares parameters for the query.

    :param row_type: type of expected result. Usually some dataclass or named tuple
    :param sql: SQL statement to execute. If None, the SQL statement must be provided
    by decorated function.

    IMPORTANT: decorated function must not be async, but after decoration it
    becomes async.

    More info in the noorm.asyncpg docstring.
    """

    def decorator(
        func: Callable[F_Spec, PrepareFuncResult | None],
    ) -> Callable[
        Concatenate[asyncpg.Connection, F_Spec], Coroutine[Any, Any, TR | None]
    ]:
        class wrapper(WrapperBase):
            async def __call__(
                self,
                conn: asyncpg.Connection,
                *args: F_Spec.args,
                **kwargs: F_Spec.kwargs,
            ) -> TR | None:
                with MetricsCollector(self._func) as mc:
                    if sql_and_params := req_sql_n_params(
                        self._func, args, kwargs, sql
                    ):
                        q_res = await conn.fetchrow(
                            sql_and_params[0], *sql_and_params[1]
                        )
                    else:
                        return None
                    if q_res is None:
                        return None
                    mc.tuples = 1
                    return row_type(**{n: v for n, v in q_res.items()})

        return wrapper(func)

    return decorator


def sql_scalar_or_none(res_type: Type[TR], sql: str | None = None):
    """
    Use this decorator to make a "scalar" SQL statement executor out of
    the function that prepares parameters for the query.

    :param res_type: type of expected result. For scalar queries it is usually `int`,
    `str`, `bool`, `datetime`, or whatever can be produced by scalar query.
    :param sql: SQL statement to execute. If None, the SQL statement must be provided
    by decorated function.

    IMPORTANT: decorated function must not be async, but after decoration it
    becomes async.

    More info in the noorm.asyncpg docstring.
    """

    def decorator(
        func: Callable[F_Spec, PrepareFuncResult | None],
    ) -> Callable[
        Concatenate[asyncpg.Connection, F_Spec], Coroutine[Any, Any, TR | None]
    ]:
        class wrapper(WrapperBase):
            async def __call__(
                self,
                conn: asyncpg.Connection,
                *args: F_Spec.args,
                **kwargs: F_Spec.kwargs,
            ) -> TR | None:
                with MetricsCollector(self._func) as mc:
                    if sql_and_params := req_sql_n_params(
                        self._func, args, kwargs, sql
                    ):
                        q_res = await conn.fetchval(
                            sql_and_params[0], *sql_and_params[1]
                        )
                        if q_res is not None:
                            mc.tuples = 1
                    else:
                        return None
                    return q_res

        return wrapper(func)

    return decorator


def sql_fetch_scalars(res_type: Type[TR], sql: str | None = None):
    """
    Use this decorator to make a "fetch scalars" SQL statement executor out of
    the function that prepares parameters for the query.

    :param row_type: type of expected result. For scalar queries it is usually `int`,
    `str`, `bool`, `datetime`, or whatever can be produced in a single-column query
    result.
    :param sql: SQL statement to execute. If None, the SQL statement must be provided
    by decorated function.

    IMPORTANT: decorated function must not be async, but after decoration it
    becomes async.

    More info in the noorm.asyncpg docstring.
    """

    def decorator(
        func: Callable[F_Spec, PrepareFuncResult | None]
    ) -> Callable[
        Concatenate[asyncpg.Connection, F_Spec], Coroutine[Any, Any, list[TR]]
    ]:
        class wrapper(WrapperBase):
            async def __call__(
                self,
                conn: asyncpg.Connection,
                *args: F_Spec.args,
                **kwargs: F_Spec.kwargs,
            ) -> list[TR]:
                with MetricsCollector(self._func) as mc:
                    if sql_and_params := req_sql_n_params(
                        self._func, args, kwargs, sql
                    ):
                        q_res = await conn.fetch(sql_and_params[0], *sql_and_params[1])
                        res: list[TR] = [r[0] for r in q_res]
                        mc.tuples = len(res)
                        return res
                    return []

        return wrapper(func)

    return decorator


@overload
def sql_execute(
    func: Callable[F_Spec, PrepareFuncResult | None]
) -> Callable[Concatenate[asyncpg.Connection, F_Spec], Coroutine[Any, Any, None]]:
    pass  # pragma: no cover


@overload
def sql_execute(
    sql: str | None = None,
) -> Callable[
    [Callable[F_Spec, Coroutine[Any, Any, None]]],
    Callable[Concatenate[asyncpg.Connection, F_Spec], Coroutine[Any, Any, None]],
]:
    pass  # pragma: no cover


def sql_execute(  # type: ignore
    sql: Callable[F_Spec, PrepareFuncResult | None] | str | None = None
):
    """
    Use this decorator to execute a statement without responding a result.

    :param sql: SQL statement to execute. If None, the SQL statement must be provided
    by decorated function.

    IMPORTANT: decorated function must not be async, but after decoration it
    becomes async.

    More info in the noorm.asyncpg docstring.
    """

    if callable(sql):
        the_func = sql
        the_sql = None
    else:
        the_func = None
        the_sql = sql

    def decorator_wrapper(sql: str | None):
        def decorator(
            func: Callable[F_Spec, PrepareFuncResult | None],
        ) -> Callable[
            Concatenate[asyncpg.Connection, F_Spec], Coroutine[Any, Any, None]
        ]:
            class wrapper(WrapperBase):
                async def __call__(
                    self,
                    conn: asyncpg.Connection,
                    *args: F_Spec.args,
                    **kwargs: F_Spec.kwargs,
                ) -> None:
                    with MetricsCollector(self._func):
                        if sql_and_params := req_sql_n_params(
                            self._func, args, kwargs, sql
                        ):
                            await conn.execute(sql_and_params[0], *sql_and_params[1])

            return wrapper(func)

        return decorator

    wrap_decorator = decorator_wrapper(the_sql)
    if callable(sql):
        return wrap_decorator(the_func)
    else:
        return wrap_decorator



"""
NoORM adapter for synchronous access to SQLite via Standard Python library.

Decorators:
- `sql_fetch_all(row_type: type, sql: str)` to fetch records as a list
- `sql_one_or_none(res_type: type, sql: str)` to fetch one record
- `sql_scalar_or_none(res_type: type, sql: str)` to fetch a scalar
- `sql_fetch_scalars(res_type: type, sql: str)` to fetch a list of scalars
- `sql_execute(sql: str)` to execute a statement
- `sql_iterate(row_type: type, sql: str)` and
  `sql_iterate_scalars(row_type: type, sql: str)` to make a query and iterate
  through results, objects or scalars respectively.
- `default_db` additional decorartor to make your DB API functions a bit easier to use.
  Check the `noorm.sqlite3.default_db` docstring for descriprion and examples.

Decorator parameters are:
- (except `sql_execute`) Expected result type. For `sql_fetch_all` and `sql_one_or_none`
  is usually a dataclass or namedtuple; for `sql_scalar_or_none` and `sql_fetch_scalars`
  is usually `int`, `str`, `bool`, `datetime`, or whatever can be produced by scalar
  query.
- SQL statement as string. If None, must be provided from the decorated function.
  Query parameters placeholders are `?` for positional and `:name` for named arguments:
  - `select rowid, username from users where username like ? and birthdate >= ?`
  - `select rowid, username from users where username like :name and birthdate >= :bdate`

After decoration the decorated function receives an open SQLite connection or a Cursor
as a first positional argument.

Decorated function should return:
- `None` if no changes to original sql statement is needed,
  and no params to be applied
- Use `params`, `query_only`, or `query_and_params` functions to return accordingly
  parameters values, SQL statement, or both together.
  - `params(*args, **kwargs)` to pass positional arguments for `?`-style placeholders
    or keyword arguments for named query parameter placeholders. SQL statement is
    original from decorator parameter.
  - `query_only(sql: str)` - only SQL query and no query parameters.
  - `query_and_params(sql: str, *args, **kwargs)` to provide both - a query and its
    parameters.

SQLite-specific features:
1. You can use `date`, `datetime`, `bool`, `Decimal` fields in query result dataclasses.
   Conversion will be done automatically.
2. You can also pass values of these types as parameters. They will be converted to
   SQLite-acceptable types automatically. Date and datetime become TEXT in ISO format,
   booleans become INTEGER, Decimal becomes TEXT.
3. Although SQLite engine does not accept collections as parameter values, here you can
   do it. That is especially useful in `WHERE ... IN (?)` conditions. Such parameter
   can be any kind if plain collection: tuple, list, set, etc., but not dict. See
   the `get_several_users` example below.

Examples:
```
from collections import namedtuple
import sqlite3

import noorm.sqlite3 as nm

# Fetch all, no parameters
@nm.sql_fetch_all(
    namedtuple("DbAllUsersResult", "id, username"),
    "select rowid as id, username from users order by rowid;",
)
def get_all_users():
    pass  # no parameters for this query

# Fetch one, the '?' style parameter
@nm.sql_one_or_none(
    namedtuple("DbOneUserResult", "username, email"),
    "select username, email from users where rowid = ?;",  # Placeholder is "?"
)
def get_one_user(rowid: int):
    return nm.params(rowid)

# Fetch scalar using named parameter
@nm.sql_scalar_or_none(
    int,
    "select count(*) from users where username like :search;"  # Named placeholder
)
def get_num_users(search: str):
    return nm.params(search=f"%{search}%")

# Fetch several users
@nm.sql_fetch_all(
    namedtuple("DbUsersResult", "id, username, email"),
    "select rowid as id, username, email from users where rowid in (:ids);",
)
def get_several_users(ids: list[int]):
    return nm.params(ids=ids)

# Insert a new record
@nm.sql_execute("insert into users(username, email) values(?, ?)")
def ins_user(username: str | None = None, email: str | None = None):
    return nm.params(username, email)

# Usage:
with sqlite3.connect("test.sqlite") as conn:  # "test.sqlite" is a DB file name
    for user in get_all_users(conn):
        print(user)
    print(f"User with id=1: {get_one_user(conn, 1)}")
    print(f"Search results by 'John': {get_num_users(conn, 'John')}")

    ins_user(conn, "Jane", "jane@example.com")
    conn.commit()  # Do not forget to commit the data manipulation!!!

```
"""  # noqa: E501

from ._sqlite3 import (
    sql_fetch_all,
    sql_iterate,
    sql_one_or_none,
    sql_scalar_or_none,
    sql_fetch_scalars,
    sql_iterate_scalars,
    sql_execute,
    default_db,
    set_default_db,
)
from noorm._db_api_2 import params, query_and_params, query_only
from noorm._common import CancelExecException

__all__ = [
    "sql_fetch_all",
    "sql_iterate",
    "sql_one_or_none",
    "sql_scalar_or_none",
    "sql_fetch_scalars",
    "sql_iterate_scalars",
    "sql_execute",
    "default_db",
    "set_default_db",
    "params",
    "query_and_params",
    "query_only",
    "CancelExecException",
]



"""
NoORM adapter for synchronous SQLAlchemy.

Decorators:
- `sql_fetch_all(row_type: type, no_commit: bool, sync_session: bool | str | None)`
  to fetch records as a list
- `sql_one_or_none(res_type: type, no_commit: bool, sync_session: bool | str | None)`
  to fetch one record
- `sql_scalar_or_none(res_type: type, no_commit: bool, sync_session: bool | str | None)`
  to fetch a scalar
- `sql_fetch_scalars(res_type: type, no_commit: bool, sync_session: bool | str | None)`
  to fetch a list of scalars
- `sql_execute(no_commit: bool, sync_session: bool | str | None)` to execute a statement
- `sql_iterate(row_type: type, no_commit: bool, sync_session: bool | str | None)` and
  `sql_iterate_scalars(row_type: type, no_commit: bool, sync_session: bool | str | None)`
  to make a query and iterate through results, objects or scalars respectively.

Decorator parameters are:
- (except `sql_execute`) Expected result type. For `sql_fetch_all` and `sql_one_or_none`
  is usually a dataclass or namedtuple; for `sql_scalar_or_none` and `sql_fetch_scalars`
  is usually `int`, `str`, `bool`, `datetime`, or whatever can be produced by scalar
  query.
- "No commit" flag. Default is False, so if executed statement is data manipulation
  (insert, update, or delete), commit will be automatically invoked. If you need
  to prevent it for some reason, set no_commit=True.
- "synchronize_session" SqlAlchemy execution option. By default is False because
  avoiding using persistent objects is one of the most crucial NoORM ideas. If no
  persistent object in a session, there is nothing to synchronise. If you really need
  different behavior, set this `sync_session` parameter to "fetch" or "evaluate", or
  pass None to turn off the execution option manipulation.

After decoration the decorated function receives an open Session as a first positional
argument.

Decorated function should return an executable statement (select, insert, update,
delete).

Examples:
```
from collections import namedtuple

import sqlalchemy as sa
import noorm.sqlalchemy_sync as nm

from db import User  # Model class for users table

# Fetch all
@nm.sql_fetch_all(namedtuple("DbAllUsersResult", "id, username, email"))
def get_all_users():
    return sa.select(User.id, User.username, User.email).order_by(User.id))

# Fetch one
@nm.sql_one_or_none(namedtuple("DbOneUserResult", "username, email"))
def get_one_user(id_: int):
    return sa.select(User.username, User.email).where(User.id == id_)

# Insert a new record
@nm.sql_execute
def ins_user(username: str | None = None, email: str | None = None):
    return sa.insert(User).values(username=username, email=email)

# Usage:
    for user in get_all_users(session):
        print(user)
    print(f"User with id=1: {get_one_user(session, 1)}")

    ins_user(session, "Jane", "jane@example.com") # Commit is done automatically
```
"""  # noqa: E501

from ._sqlalchemy_sync import (
    sql_fetch_all,
    sql_iterate,
    sql_one_or_none,
    sql_scalar_or_none,
    sql_fetch_scalars,
    sql_iterate_scalars,
    sql_execute,
)
from noorm._common import CancelExecException

__all__ = [
    "sql_fetch_all",
    "sql_iterate",
    "sql_one_or_none",
    "sql_scalar_or_none",
    "sql_fetch_scalars",
    "sql_iterate_scalars",
    "sql_execute",
    "CancelExecException",
]



from typing import Type, Callable, Generator, ParamSpec, TypeVar, Concatenate, overload
import sqlite3

from .._common import WrapperBase
from .._sqlite_common import (
    make_decoder as _make_decoder,
    make_scalar_decoder as _make_scalar_decoder,
    sqlite_sql_n_params,
)
from .._db_api_2 import PrepareFuncResult
from ..registry import MetricsCollector

F_Spec = ParamSpec("F_Spec")
F_Return = TypeVar("F_Return")
TR = TypeVar("TR")
ConnectionOrCursor = sqlite3.Connection | sqlite3.Cursor


def sql_fetch_all(row_type: Type[TR], sql: str | None = None):
    """
    Use this decorator to make a "fetch all" SQL statement executor out of
    the function that prepares parameters for the query.

    :param row_type: type of expected result. Usually some dataclass or named tuple
    :param sql: SQL statement to execute. If None, the SQL statement must be provided
    by decorated function.

    More info in the noorm.sqlite3 docstring.
    """

    def decorator(
        func: Callable[F_Spec, PrepareFuncResult | None]
    ) -> Callable[Concatenate[ConnectionOrCursor, F_Spec], list[TR]]:
        decoder = _make_decoder(row_type)

        class wrapper(WrapperBase):
            def __call__(
                self,
                conn: ConnectionOrCursor,
                *args: F_Spec.args,
                **kwargs: F_Spec.kwargs,
            ) -> list[TR]:
                with MetricsCollector(self._func) as mc:
                    if sql_and_params := sqlite_sql_n_params(
                        self._func, args, kwargs, sql
                    ):
                        q_res = conn.execute(*sql_and_params)
                        col_names = tuple(el[0] for el in q_res.description)
                        res: list[TR] = [
                            row_type(**decoder({n: v for n, v in zip(col_names, r)}))
                            for r in q_res
                        ]
                        mc.tuples = len(res)
                        return res
                    return []

        return wrapper(func)

    return decorator


def sql_iterate(row_type: Type[TR], sql: str | None = None):
    """
    Use this decorator to make a query and iterate through results. Be careful with
    this feature and, if possible, use `sql_fetch_all` instead, because
    `sql_fetch_all` gives you less possibilites to shoot your leg.

    :param row_type: type of expected result. Usually some dataclass or named tuple
    :param sql: SQL statement to execute. If None, the SQL statement must be provided
    by decorated function.

    More info in the noorm.sqlite3 docstring.
    """

    def decorator(
        func: Callable[F_Spec, PrepareFuncResult | None]
    ) -> Callable[Concatenate[sqlite3.Connection, F_Spec], Generator[TR, None, None]]:
        decoder = _make_decoder(row_type)

        class wrapper(WrapperBase):
            def __call__(
                self,
                conn: sqlite3.Connection,
                *args: F_Spec.args,
                **kwargs: F_Spec.kwargs,
            ) -> Generator[TR, None, None]:
                with MetricsCollector(self._func) as mc:
                    if sql_and_params := sqlite_sql_n_params(
                        self._func, args, kwargs, sql
                    ):
                        cur = conn.cursor()
                        q_res = cur.execute(*sql_and_params)
                        col_names = tuple(el[0] for el in q_res.description)
                        is_first_row = True
                        for r in q_res:
                            if is_first_row:
                                mc.finish(None)
                                is_first_row = False
                            yield row_type(
                                **decoder({n: v for n, v in zip(col_names, r)})
                            )

        return wrapper(func)

    return decorator


def sql_one_or_none(row_type: Type[TR], sql: str | None = None):
    """
    Use this decorator to make a "one or none" SQL statement executor out of
    the function that prepares parameters for the query.

    :param row_type: type of expected result. Usually some dataclass or named tuple
    :param sql: SQL statement to execute. If None, the SQL statement must be provided
    by decorated function.

    More info in the noorm.sqlite3 docstring.
    """

    def decorator(
        func: Callable[F_Spec, PrepareFuncResult | None]
    ) -> Callable[Concatenate[ConnectionOrCursor, F_Spec], TR | None]:
        decoder = _make_decoder(row_type)

        class wrapper(WrapperBase):
            def __call__(
                self,
                conn: ConnectionOrCursor,
                *args: F_Spec.args,
                **kwargs: F_Spec.kwargs,
            ) -> TR | None:
                with MetricsCollector(self._func) as mc:
                    if sql_and_params := sqlite_sql_n_params(
                        self._func, args, kwargs, sql
                    ):
                        q_res = conn.execute(*sql_and_params)
                        col_names = tuple(el[0] for el in q_res.description)
                        for row in q_res:
                            mc.tuples = 1
                            return row_type(
                                **decoder({n: v for n, v in zip(col_names, row)})
                            )
                    return None

        return wrapper(func)

    return decorator


def sql_scalar_or_none(res_type: Type[TR], sql: str | None = None):
    """
    Use this decorator to make a "scalar or none" SQL statement executor out of
    the function that prepares parameters for the query.

    :param row_type: type of expected result. For scalar queries it is usually `int`,
    `str`, `bool`, `datetime`, or whatever can be produced by scalar query. Use Any to
    return a value without conversion.
    :param sql: SQL statement to execute. If None, the SQL statement must be provided
    by decorated function.

    More info in the noorm.sqlite3 docstring.
    """

    def decorator(
        func: Callable[F_Spec, PrepareFuncResult | None]
    ) -> Callable[Concatenate[ConnectionOrCursor, F_Spec], TR | None]:
        decoder = _make_scalar_decoder(res_type)

        class wrapper(WrapperBase):
            def __call__(
                self,
                conn: ConnectionOrCursor,
                *args: F_Spec.args,
                **kwargs: F_Spec.kwargs,
            ) -> TR | None:
                with MetricsCollector(self._func) as mc:
                    if sql_and_params := sqlite_sql_n_params(
                        self._func, args, kwargs, sql
                    ):
                        q_res = conn.execute(*sql_and_params)
                        for row in q_res:
                            mc.tuples = 1
                            return decoder(row[0])
                    return None

        return wrapper(func)

    return decorator


def sql_fetch_scalars(res_type: Type[TR], sql: str | None = None):
    """
    Use this decorator to make a "fetch scalars" SQL statement executor out of
    the function that prepares parameters for the query.

    :param res_type: type of expected result. For scalar queries it is usually `int`,
    `str`, `bool`, `datetime`, or whatever can be produced in a single-column query
    result. Use Any to return values without conversion.
    :param sql: SQL statement to execute. If None, the SQL statement must be provided
    by decorated function.

    More info in the noorm.sqlite3 docstring.
    """

    def decorator(
        func: Callable[F_Spec, PrepareFuncResult | None]
    ) -> Callable[Concatenate[ConnectionOrCursor, F_Spec], list[TR]]:
        decoder = _make_scalar_decoder(res_type)

        class wrapper(WrapperBase):
            def __call__(
                self,
                conn: ConnectionOrCursor,
                *args: F_Spec.args,
                **kwargs: F_Spec.kwargs,
            ) -> list[TR]:
                with MetricsCollector(self._func) as mc:
                    if sql_and_params := sqlite_sql_n_params(
                        self._func, args, kwargs, sql
                    ):
                        q_res = conn.execute(*sql_and_params)
                        res = [decoder(row[0]) for row in q_res]
                        mc.tuples = len(res)
                        return res
                    return []

        return wrapper(func)

    return decorator


def sql_iterate_scalars(res_type: Type[TR], sql: str | None = None):
    """
    Use this decorator to make a query and iterate through scalar results. Be careful
    with this feature and, if possible, use `sql_fetch_scalars` instead, because
    `sql_fetch_scalars` gives you less possibilites to shoot your leg.

    :param res_type: type of expected result. For scalar queries it is usually `int`,
    `str`, `bool`, `datetime`, or whatever can be produced in a single-column query
    result. Use Any to return values without conversion.
    :param sql: SQL statement to execute. If None, the SQL statement must be provided
    by decorated function.

    More info in the noorm.sqlite3 docstring.
    """

    def decorator(
        func: Callable[F_Spec, PrepareFuncResult | None]
    ) -> Callable[Concatenate[sqlite3.Connection, F_Spec], Generator[TR, None, None]]:
        decoder = _make_scalar_decoder(res_type)

        class wrapper(WrapperBase):
            def __call__(
                self,
                conn: sqlite3.Connection,
                *args: F_Spec.args,
                **kwargs: F_Spec.kwargs,
            ) -> Generator[TR, None, None]:
                with MetricsCollector(self._func) as mc:
                    if sql_and_params := sqlite_sql_n_params(
                        self._func, args, kwargs, sql
                    ):
                        cur = conn.cursor()
                        q_res = cur.execute(*sql_and_params)
                        is_first_row = True
                        for r in q_res:
                            if is_first_row:
                                mc.finish(None)
                                is_first_row = False
                            yield decoder(r[0])

        return wrapper(func)

    return decorator


@overload
def sql_execute(
    func: Callable[F_Spec, PrepareFuncResult | None]
) -> Callable[Concatenate[ConnectionOrCursor, F_Spec], None]:
    pass  # pragma: no cover


@overload
def sql_execute(
    sql: str | None = None,
) -> Callable[
    [Callable[F_Spec, None]], Callable[Concatenate[ConnectionOrCursor, F_Spec], None]
]:
    pass  # pragma: no cover


def sql_execute(  # type: ignore
    sql: Callable[F_Spec, PrepareFuncResult | None] | str | None = None,
):
    """
    Use this decorator to execute a statement without responding a result.

    :param sql: SQL statement to execute. If None, the SQL statement must be provided
    by decorated function.

    More info in the noorm.sqlite3 docstring.
    """

    if callable(sql):
        the_func = sql
        the_sql = None
    else:
        the_func = None
        the_sql = sql

    def decorator_wrapper(sql: str | None):
        def decorator(
            func: Callable[F_Spec, PrepareFuncResult | None]
        ) -> Callable[Concatenate[ConnectionOrCursor, F_Spec], None]:
            class wrapper(WrapperBase):
                def __call__(
                    self,
                    conn: ConnectionOrCursor,
                    *args: F_Spec.args,
                    **kwargs: F_Spec.kwargs,
                ) -> None:
                    with MetricsCollector(self._func):
                        if sql_and_params := sqlite_sql_n_params(
                            self._func, args, kwargs, sql
                        ):
                            conn.execute(*sql_and_params)

            return wrapper(func)

        return decorator

    wrap_decorator = decorator_wrapper(the_sql)
    if callable(sql):
        return wrap_decorator(the_func)
    else:
        return wrap_decorator


class set_default_db:
    """
    Use `nm.set_default_db(your_connection)` as a function or as a context manager
    to set your "default" DB connection before first usage of function decorated with
    the `@nm.default_db` decorator.
    """

    _conns: list[sqlite3.Connection | None] = []

    def __init__(self, conn: sqlite3.Connection) -> None:
        if not self._conns:
            self._conns.append(None)
        self.prev = self._conns[-1]
        self._conns[-1] = conn

    def __enter__(self) -> None:
        curr = self._conns[-1]
        self._conns[-1] = self.prev
        self._conns.append(curr)

    def __exit__(self, exc_type, exc_value, traceback):
        self._conns.pop()


def default_db(
    func: Callable[Concatenate[ConnectionOrCursor, F_Spec], F_Return]
) -> Callable[F_Spec, F_Return]:
    """
    The `@nm.default_db` decorator makes your DB API functions easier to use by
    removing the first mandatory `ConnectionOrCursor` argument.

    Use this decorator before other `@nm.sql_...` decorators.

    Example:
    ```
    import sqlite3
    import noorm.sqlite3 as nm

    @nm.default_db
    @nm.sql_scalar_or_none(int, "select count(*) from users")
    def get_users_count():
        pass

    with sqlite3.connect("my_db.sqlite") as conn, nm.set_default_db(conn):
        users_count = get_users_count()  # <<< Consider no "conn" parameter
        print(f"{users_count=}")
    ```
    Without `nm.set_default_db(conn)` any call to `get_users_count` would fail with a
    runtime error "default_db is not set".

    You can use `nm.set_default_db` as a function. This code also works:
    ```
    conn = sqlite3.connect("my_db.sqlite")
    nm.set_default_db(conn)
    users_count = get_users_count()
    ```
    The `nm.set_default_db` context can be nested:
    ```
    conn1 = sqlite3.connect("my_db_1.sqlite")
    conn2 = sqlite3.connect("my_db_2.sqlite")
    with nm.set_default_db(conn1):
        print(f"Users count in my_db_1: {get_users_count()}")
        with nm.set_default_db(conn2):  # <<< Temporary set conn2 as a default_db
            print(f"Users count in my_db_2: {get_users_count()}")
        print(f"Check again users count in my_db_1: {get_users_count()}")
    ```
    """

    def wrapper(*args: F_Spec.args, **kwargs: F_Spec.kwargs) -> F_Return:
        if not set_default_db._conns or (conn := set_default_db._conns[-1]) is None:
            raise RuntimeError(
                "default_db is not set. Use nm.set_default_db() to set default DB "
                "connection before the first usage of function decorated "
                "with @nm.default_db"
            )
        return func(conn, *args, **kwargs)

    return wrapper



from typing import Type, Any, Callable, Collection, Mapping
from dataclasses import is_dataclass, fields
from decimal import Decimal
from datetime import date, datetime
from functools import partial
import re

from ._db_api_2 import req_sql_n_params


def _decode_date(val: str | date | datetime | None) -> date | None:
    if val is None:
        return None
    if isinstance(val, datetime):
        return val.date()
    if isinstance(val, date):
        return val
    return date.fromisoformat(val)


def _decode_datetime(val: str | date | datetime | None) -> datetime | None:
    if val is None:
        return None
    if isinstance(val, datetime):
        return val
    if isinstance(val, date):
        return datetime(val.year, val.month, val.day)
    return datetime.fromisoformat(val)


def _as_is(val: Any) -> Any:
    return val


def _as_is_val(type_: Type, val: Any) -> Any:
    if val is None:
        return None
    if isinstance(val, type_):
        return val
    return type_(val)


def make_scalar_decoder(type_: Type) -> Callable[[Any], Any]:
    if type_ is Any:
        return _as_is
    for ttype in (int, float, str, bool, Decimal):
        if issubclass(ttype, type_):
            return partial(_as_is_val, ttype)
    if issubclass(date, type_):
        return _decode_date
    if issubclass(datetime, type_):
        return _decode_datetime
    return _as_is


def make_decoder(row_type: type) -> Callable[[dict], dict]:
    if is_dataclass(row_type):
        fields_mapper: dict[str, Callable] = {
            fld.name: make_scalar_decoder(fld.type) for fld in fields(row_type)
        }

        def _dataclass_decoder(inp: dict) -> dict:
            return {k: fields_mapper[k](v) for k, v in inp.items()}

        return _dataclass_decoder

    return _as_is


def _encode_val(val: Any) -> Any:
    if isinstance(val, bool):
        return int(val)
    elif isinstance(val, int | str | None | float | bytes | bytearray):
        return val
    elif isinstance(val, datetime):
        return val.isoformat(sep=" ")
    elif isinstance(val, date):
        return val.isoformat()
    elif isinstance(val, Decimal):
        return str(val)
    elif getattr(type(val), "__conform__", None) is not None:
        return val
    return ...


_SQL_SPLITTER = re.compile(r"'.*?'|\".*?\"|--.*?\n")


def _propagate_named_params(list_param_sizes: dict[str, int], code: str) -> str:
    for pname, size in list_param_sizes.items():
        code_parts: list[str] = []
        pname_len = len(pname)
        cleaned_pname = pname[1:]
        while True:
            pos = code.find(pname)
            if pos == -1:
                code = "".join(code_parts) + code
                break
            if pos > 0:
                code_parts.append(code[:pos])
                code = code[pos:]
            next_char = code[pname_len : pname_len + 1]
            if next_char and next_char.isalnum() or next_char == "_":
                code_parts.append(code[:pname_len])
            else:
                code_parts.append(
                    ", ".join(f":__{cleaned_pname}_{i}" for i in range(size))
                )
            code = code[pname_len:]
    return code


def _propagate_positional_params(
    list_param_sizes: dict[int, int], next_idx_var: list[int], code: str
) -> str:
    split_code = code.split("?")
    code_parts: list[str] = []
    for qmark_idx in range(len(split_code) - 1):
        code_parts.append(split_code[qmark_idx])
        if next_idx_var[0] in list_param_sizes:
            code_parts.append(
                ", ".join("?" for _ in range(list_param_sizes[next_idx_var[0]]))
            )
        else:
            code_parts.append("?")
        next_idx_var[0] += 1
    return "".join(code_parts) + split_code[-1]


def _adjust_sql(sql: str, propagate_func: Callable) -> str:
    res_sqls = []
    last_pos = 0
    for mtch in _SQL_SPLITTER.finditer(sql):
        span = mtch.span()
        code = propagate_func(sql[last_pos : span[0]])
        res_sqls.append(code)
        res_sqls.append(sql[span[0] : span[1]])
        last_pos = span[1]
    if last_pos < len(sql):
        code = propagate_func(sql[last_pos:])
        res_sqls.append(code)
    return "".join(res_sqls)


def _prepare_named_params(sql: str, params: dict) -> tuple[str, dict]:
    list_param_sizes: dict[str, int] = {}
    res_params = {}
    for pname, pvalue in params.items():
        val = _encode_val(pvalue)
        if val is not Ellipsis:
            res_params[pname] = val
        else:
            if isinstance(pvalue, Collection) and not isinstance(pvalue, Mapping):
                # Propagate params
                list_param_sizes[":" + pname] = len(pvalue)
                for idx, pval in enumerate(pvalue):
                    val = _encode_val(pval)
                    if val is Ellipsis:
                        raise TypeError(
                            f"Wrong type of parameter {pname}[{idx}]: "
                            f"{type(pval).__name__}"
                        )
                    else:
                        res_params[f"__{pname}_{idx}"] = val
            else:
                raise TypeError(
                    f"Wrong type of parameter {pname}: {type(pvalue).__name__}"
                )
    if not list_param_sizes:
        res_sql = sql
    else:
        res_sql = _adjust_sql(sql, partial(_propagate_named_params, list_param_sizes))
    return res_sql, res_params


def _prepare_positional_params(sql: str, params: tuple) -> tuple[str, tuple]:
    list_param_sizes: dict[int, int] = {}
    res_params = []
    for pidx, pvalue in enumerate(params):
        val = _encode_val(pvalue)
        if val is not Ellipsis:
            res_params.append(val)
        else:
            if isinstance(pvalue, Collection) and not isinstance(pvalue, Mapping):
                # Propagate params
                list_param_sizes[pidx] = len(pvalue)
                for idx, pval in enumerate(pvalue):
                    val = _encode_val(pval)
                    if val is Ellipsis:
                        raise TypeError(
                            f"Wrong type of parameter {pidx}[{idx}]: "
                            f"{type(pval).__name__}"
                        )
                    else:
                        res_params.append(val)
            else:
                raise TypeError(
                    f"Wrong type of parameter {pidx}: {type(pvalue).__name__}"
                )
    if not list_param_sizes:
        res_sql = sql
    else:
        next_idx_var = [0]
        res_sql = _adjust_sql(
            sql, partial(_propagate_positional_params, list_param_sizes, next_idx_var)
        )

    return res_sql, tuple(res_params)


def sqlite_sql_n_params(
    func, f_args, f_kwargs, default_sql: str | None
) -> tuple[str, dict | tuple] | None:
    sql_n_params = req_sql_n_params(func, f_args, f_kwargs, default_sql)
    if sql_n_params is None:
        return None
    sql, params = sql_n_params
    if isinstance(params, dict):
        sql, params = _prepare_named_params(sql, params)
    else:
        sql, params = _prepare_positional_params(sql, params)
    return sql, params



from sqlalchemy.sql import Executable, Select

from ._common import CancelExecException


def req_sql_n_params(
    func, f_args, f_kwargs, sync_session: bool | str | None
) -> Executable | None:
    try:
        sql: Executable = func(*f_args, **f_kwargs)
    except CancelExecException:
        return None
    if isinstance(sql, Executable):
        if not isinstance(sql, Select) and sync_session is not None:
            sql = sql.execution_options(synchronize_session=sync_session)
        return sql
    raise TypeError(
        f"Function {func.__name__} returned {type(sql).__name__} (expected Executable)"
    )



from .registry import get_registry


class CancelExecException(Exception):
    """Do not execute any query and respond empty result"""


class WrapperBase:
    def __init__(self, func) -> None:
        self._func = func
        get_registry().register(func)

    def unwrapped(self, *args, **kwargs):
        return self._func(*args, **kwargs)



import logging
from typing import Callable
from dataclasses import dataclass, field
from collections import defaultdict
from functools import lru_cache
from time import perf_counter
from multiprocessing import current_process
from threading import Thread
import socket
import os
import pickle


@dataclass
class Stat:
    calls: int = 0
    duration: float = 0.0
    tuples: int = 0
    fails: int = 0
    fails_by_error: dict[str, int] = field(default_factory=dict)


@dataclass(slots=True, frozen=True)
class FuncCallEvent:
    func_name: str
    duration: float | None
    tuples: int
    error: str | None


class Registry:
    def __init__(self) -> None:
        self.name_by_func: dict[Callable, str] = {}
        self.func_names: set[str] = set()
        self.stat_by_name: defaultdict[str, Stat] = defaultdict(Stat)
        self._event_listeners: list[Callable[[FuncCallEvent], None]] = []
        self._mp_listener_started = False
        self._is_main_process = current_process().name == "MainProcess"
        self._mp_socket: socket.socket | None = None
        self._mp_port = 0
        if not self._is_main_process:
            self._mp_socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
            reg_port = int(os.getenv("NOORM_REGISTRY_PORT") or "0")
            if reg_port == 0:
                logging.warning(
                    "Noorm works in a child process, but 'init_multiprocess_registry' "
                    "on a main process was not invoked. All statistics on this child "
                    "process will be lost."
                )
            self._mp_port = reg_port

    def _main_process_socket_listener(self) -> None:
        while self._mp_socket is not None:
            bin_data, _ = self._mp_socket.recvfrom(4096)
            try:
                data = pickle.loads(bin_data)
                # data is a func name or a FuncCallEvent
                if isinstance(data, str):
                    if data == "#stop#":
                        self._mp_socket.close()
                        self._mp_socket = None
                        return
                    self.func_names.add(data)
                else:
                    assert isinstance(data, FuncCallEvent)
                    self.on_event(data)
            except Exception:
                pass  # do nothing

    def init_multiprocess_registry(self) -> None:
        if not self._mp_listener_started:
            if not self._is_main_process:
                logging.warning(
                    "Cannot init multiprocess noorm registry in a child process"
                )
                return
            self._mp_socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
            self._mp_socket.bind(("localhost", 0))
            self._mp_port = self._mp_socket.getsockname()[1]
            os.environ["NOORM_REGISTRY_PORT"] = str(self._mp_port)
            self._mp_listener_started = True
            Thread(target=self._main_process_socket_listener, daemon=True).start()

    def close_multiprocess_registry(self) -> None:
        """Mainly for testing purposes"""
        if self._mp_listener_started:
            socket.socket(socket.AF_INET, socket.SOCK_DGRAM).sendto(
                pickle.dumps("#stop#"), ("localhost", self._mp_port)
            )
            del os.environ["NOORM_REGISTRY_PORT"]
            self._mp_listener_started = False

    def register(self, func: Callable) -> str:
        func_name = func.__module__ + "." + func.__qualname__
        self.func_names.add(func_name)
        self.name_by_func[func] = func_name
        if not self._is_main_process and self._mp_port and self._mp_socket is not None:
            try:
                self._mp_socket.sendto(
                    pickle.dumps(func_name), ("localhost", self._mp_port)
                )
            except Exception:
                pass  # do nothing
        return func_name

    def add_event_listener(self, callback: Callable[[FuncCallEvent], None]) -> None:
        self._event_listeners.append(callback)

    def on_event(self, event: FuncCallEvent) -> None:
        if not self._is_main_process:
            if self._mp_port and self._mp_socket is not None:
                try:
                    self._mp_socket.sendto(
                        pickle.dumps(event), ("localhost", self._mp_port)
                    )
                except Exception:
                    pass  # do nothing
        else:
            stat = self.stat_by_name[event.func_name]
            stat.calls += 1
            stat.duration += event.duration or 0
            stat.tuples += event.tuples
            if event.error:
                stat.fails += 1
                stat.fails_by_error[event.error] = (
                    stat.fails_by_error.get(event.error, 0) + 1
                )

            for callback in self._event_listeners:
                try:
                    callback(event)
                except Exception:
                    pass  # Just ignore all the fails in external listeners

    def clear_stat(self) -> None:
        self.stat_by_name.clear()


class MetricsCollector:
    def __init__(self, func: Callable) -> None:
        self.registry = get_registry()
        self.func = func
        self.tuples = 0
        self.start_time: float | None = None

    def finish(self, exc_type: type | None) -> None:
        if self.start_time is not None:
            duration = perf_counter() - self.start_time
            self.start_time = None
            self.registry.on_event(
                FuncCallEvent(
                    (
                        self.registry.name_by_func.get(self.func)
                        or self.registry.register(self.func)
                    ),
                    duration,
                    self.tuples,
                    None if exc_type is None else exc_type.__name__,
                )
            )

    def __enter__(self):
        self.start_time = perf_counter()
        return self

    def __exit__(self, exc_type: type | None, exc_value, traceback) -> None:
        self.finish(exc_type)


@lru_cache
def _get_registry(pid):
    return Registry()


def get_registry():
    return _get_registry(current_process().pid)



"""
Implements DB_API_2-compliant 'params', 'query_and_params', and 'query_only'
functions.
"""

from ._common import CancelExecException


class PrepareFuncResult:
    def __init__(self, sql: str | None, params: tuple | dict | None) -> None:
        self.sql = sql
        self.params = params


def _bake_params(args: tuple, kwargs: dict) -> tuple | dict | None:
    if args and kwargs:
        raise ValueError(
            "Only positional OR keyword arguments are allowed here, not both"
        )
    if kwargs:
        return kwargs
    elif args:
        return args
    return None


def params(*args, **kwargs):
    return PrepareFuncResult(None, _bake_params(args, kwargs))


def query_and_params(sql: str, *args, **kwargs):
    return PrepareFuncResult(sql, _bake_params(args, kwargs))


def query_only(sql: str):
    return PrepareFuncResult(sql, None)


def req_sql_n_params(
    func, f_args, f_kwargs, default_sql: str | None
) -> tuple[str, dict | tuple] | None:
    try:
        sql_n_params: PrepareFuncResult | None = func(*f_args, **f_kwargs)
    except CancelExecException:
        return None

    if sql_n_params is not None:
        if not isinstance(sql_n_params, PrepareFuncResult):
            raise TypeError(
                f"Function {func.__name__} returned {type(sql_n_params).__name__} "
                "(expected None or result of 'params', 'query_and_params', or "
                "'query_only' functions)"
            )
        sql = sql_n_params.sql
        params = sql_n_params.params
        if sql is None:
            sql = default_sql
        if params is None:
            params = tuple()
    else:
        sql = default_sql
        params = tuple()
    if sql is None:
        raise RuntimeError(
            f"Function {func.__name__} did not return an SQL statement "
            "in its result. When SQL statement is not provided in decorator params, "
            "it should be returned by the function through the 'query_only' or "
            "'query_and_params' function."
        )
    return sql, params



"""
NoORM (Not Only ORM) helpers for asynchronous sqlalchemy
"""

from typing import Type, Callable, Coroutine, ParamSpec, TypeVar, Any, overload
from typing import Concatenate, AsyncGenerator

from sqlalchemy.sql import Executable, Select
from sqlalchemy.ext.asyncio import AsyncSession

from .._sqlalchemy_common import req_sql_n_params
from .._common import WrapperBase
from ..registry import MetricsCollector

F_Spec = ParamSpec("F_Spec")
TR = TypeVar("TR")


async def _commit_if_needed(
    session: AsyncSession, sql_stmt: Executable, no_commit: bool
):
    if not isinstance(sql_stmt, Select) and not no_commit:
        await session.commit()


def sql_fetch_all(
    row_type: Type[TR], no_commit: bool = False, sync_session: bool | str | None = False
):
    """
    Use this decorator to make `.all()` queries.

    :param row_type: type of expected result. Usually some dataclass or named tuple
    :param no_commit: set to False to prevent commit after the DML execution.
    :param sync_session: execution option `synchronize_session`. Default False.

    IMPORTANT: decorated function must not be async, but after decoration it
    becomes async.

    More info in the noorm.sqlalchemy_async docstring.
    """

    def decorator(
        func: Callable[F_Spec, Executable]
    ) -> Callable[Concatenate[AsyncSession, F_Spec], Coroutine[Any, Any, list[TR]]]:
        class wrapper(WrapperBase):
            async def __call__(
                self, session: AsyncSession, *args: F_Spec.args, **kwargs: F_Spec.kwargs
            ) -> list[TR]:
                with MetricsCollector(self._func) as mc:
                    if (
                        sql_stmt := req_sql_n_params(
                            self._func, args, kwargs, sync_session
                        )
                    ) is not None:
                        q_res = await session.execute(sql_stmt)
                        res: list[TR] = [
                            row_type(**{n: v for n, v in r._asdict().items()})
                            for r in q_res
                        ]
                        await _commit_if_needed(session, sql_stmt, no_commit)
                        mc.tuples = len(res)
                        return res
                    return []

        return wrapper(func)

    return decorator


def sql_iterate(
    row_type: Type[TR], no_commit: bool = False, sync_session: bool | str | None = False
):
    """
    Use this decorator to make a query and iterate through results. Be careful with
    this feature and, if possible, use `sql_fetch_all` instead, because
    `sql_fetch_all` gives you less possibilites to shoot your leg.

    :param row_type: type of expected result. Usually some dataclass or named tuple
    :param no_commit: set to False to prevent commit after the DML execution.
    :param sync_session: execution option `synchronize_session`. Default False.

    IMPORTANT: decorated function must not be async, but after decoration it
    becomes async.

    More info in the noorm.sqlalchemy_async docstring.
    """

    def decorator(
        func: Callable[F_Spec, Executable]
    ) -> Callable[Concatenate[AsyncSession, F_Spec], AsyncGenerator[TR, None]]:
        class wrapper(WrapperBase):
            async def __call__(
                self, session: AsyncSession, *args: F_Spec.args, **kwargs: F_Spec.kwargs
            ) -> AsyncGenerator[TR, None]:
                with MetricsCollector(self._func) as mc:
                    if (
                        sql_stmt := req_sql_n_params(
                            self._func, args, kwargs, sync_session
                        )
                    ) is not None:
                        q_res = await session.execute(sql_stmt)
                        is_first_row = True
                        for r in q_res:
                            if is_first_row:
                                mc.finish(None)
                                is_first_row = False
                            yield row_type(**{n: v for n, v in r._asdict().items()})
                        await _commit_if_needed(session, sql_stmt, no_commit)

        return wrapper(func)

    return decorator


def sql_one_or_none(
    row_type: Type[TR], no_commit: bool = False, sync_session: bool | str | None = False
):
    """
    Use this decorator to make `.one_or_none()` queries.

    :param row_type: type of expected result. Usually some dataclass or named tuple
    :param no_commit: set to False to prevent commit after the DML execution.
    :param sync_session: execution option `synchronize_session`. Default False.

    IMPORTANT: decorated function must not be async, but after decoration it
    becomes async.

    More info in the noorm.sqlalchemy_async docstring.
    """

    def decorator(
        func: Callable[F_Spec, Executable],
    ) -> Callable[Concatenate[AsyncSession, F_Spec], Coroutine[Any, Any, TR | None]]:
        class wrapper(WrapperBase):
            async def __call__(
                self, session: AsyncSession, *args: F_Spec.args, **kwargs: F_Spec.kwargs
            ) -> TR | None:
                with MetricsCollector(self._func) as mc:
                    if (
                        sql_stmt := req_sql_n_params(
                            self._func, args, kwargs, sync_session
                        )
                    ) is not None:
                        q_res = (await session.execute(sql_stmt)).one_or_none()
                        await _commit_if_needed(session, sql_stmt, no_commit)
                        if q_res is None:
                            return None
                        mc.tuples = 1
                        return row_type(**{n: v for n, v in q_res._asdict().items()})
                    return None

        return wrapper(func)

    return decorator


def sql_scalar_or_none(
    res_type: Type[TR], no_commit: bool = False, sync_session: bool | str | None = False
):
    """
    Use this decorator to make a "scalar" SQL statement executor out of
    the function that prepares parameters for the query

    :param res_type: type of expected result. For scalar queries it is usually `int`,
    `str`, `bool`, `datetime`, or whatever can be produced by scalar query.
    :param no_commit: set to False to prevent commit after the DML execution.
    :param sync_session: execution option `synchronize_session`. Default False.

    IMPORTANT: decorated function must not be async, but after decoration it
    becomes async.

    More info in the noorm.sqlalchemy_async docstring.
    """

    def decorator(
        func: Callable[F_Spec, Executable],
    ) -> Callable[Concatenate[AsyncSession, F_Spec], Coroutine[Any, Any, TR | None]]:
        class wrapper(WrapperBase):
            async def __call__(
                self, session: AsyncSession, *args: F_Spec.args, **kwargs: F_Spec.kwargs
            ) -> TR | None:
                with MetricsCollector(self._func) as mc:
                    if (
                        sql_stmt := req_sql_n_params(
                            self._func, args, kwargs, sync_session
                        )
                    ) is not None:
                        q_res = (await session.execute(sql_stmt)).scalar_one_or_none()
                        await _commit_if_needed(session, sql_stmt, no_commit)
                        if q_res is not None:
                            mc.tuples = 1
                        return q_res
                    return None

        return wrapper(func)

    return decorator


def sql_fetch_scalars(
    res_type: Type[TR], no_commit: bool = False, sync_session: bool | str | None = False
):
    """
    Use this decorator to make a "scalars" SQL statement executor out of
    the function that prepares parameters for the query

    :param res_type: type of expected result. For scalar queries it is usually `int`,
    `str`, `bool`, `datetime`, or whatever can be produced by scalar query.
    :param no_commit: set to False to prevent commit after the DML execution.
    :param sync_session: execution option `synchronize_session`. Default False.

    IMPORTANT: decorated function must not be async, but after decoration it
    becomes async.

    More info in the noorm.sqlalchemy_async docstring.
    """

    def decorator(
        func: Callable[F_Spec, Executable],
    ) -> Callable[Concatenate[AsyncSession, F_Spec], Coroutine[Any, Any, list[TR]]]:
        class wrapper(WrapperBase):
            async def __call__(
                self, session: AsyncSession, *args: F_Spec.args, **kwargs: F_Spec.kwargs
            ) -> list[TR]:
                with MetricsCollector(self._func) as mc:
                    if (
                        sql_stmt := req_sql_n_params(
                            self._func, args, kwargs, sync_session
                        )
                    ) is not None:
                        q_res = (await session.execute(sql_stmt)).scalars()
                        res = [el for el in q_res]
                        await _commit_if_needed(session, sql_stmt, no_commit)
                        mc.tuples = len(res)
                        return res
                    return []

        return wrapper(func)

    return decorator


def sql_iterate_scalars(
    res_type: Type[TR], no_commit: bool = False, sync_session: bool | str | None = False
):
    """
    Use this decorator to make a query and iterate through scalar results. Be careful
    with this feature and, if possible, use `sql_fetch_scalars` instead, because
    `sql_fetch_scalars` gives you less possibilites to shoot your leg.

    :param res_type: type of expected result. For scalar queries it is usually `int`,
    `str`, `bool`, `datetime`, or whatever can be produced by scalar query.
    :param no_commit: set to False to prevent commit after the DML execution.
    :param sync_session: execution option `synchronize_session`. Default False.

    IMPORTANT: decorated function must not be async, but after decoration it
    becomes async.

    More info in the noorm.sqlalchemy_async docstring.
    """

    def decorator(
        func: Callable[F_Spec, Executable]
    ) -> Callable[Concatenate[AsyncSession, F_Spec], AsyncGenerator[TR, None]]:
        class wrapper(WrapperBase):
            async def __call__(
                self, session: AsyncSession, *args: F_Spec.args, **kwargs: F_Spec.kwargs
            ) -> AsyncGenerator[TR, None]:
                with MetricsCollector(self._func) as mc:
                    if (
                        sql_stmt := req_sql_n_params(
                            self._func, args, kwargs, sync_session
                        )
                    ) is not None:
                        q_res = await session.scalars(sql_stmt)
                        is_first_row = True
                        for r in q_res:
                            if is_first_row:
                                mc.finish(None)
                                is_first_row = False
                            yield r
                        await _commit_if_needed(session, sql_stmt, no_commit)

        return wrapper(func)

    return decorator


@overload
def sql_execute(
    func: Callable[F_Spec, Executable]
) -> Callable[Concatenate[AsyncSession, F_Spec], Coroutine[Any, Any, None]]:
    pass  # pragma: no cover


@overload
def sql_execute(
    no_commit: bool = False, sync_session: bool | str | None = False
) -> Callable[
    [Callable[F_Spec, Coroutine[Any, Any, None]]],
    Callable[Concatenate[AsyncSession, F_Spec], Coroutine[Any, Any, None]],
]:
    pass  # pragma: no cover


def sql_execute(  # type: ignore
    func: Callable[F_Spec, Executable] | None = None,
    no_commit: bool = False,
    sync_session: bool | str | None = False,
):
    """
    Use this decorator to execute a statement without responding a result.

    :param no_commit: set to False to prevent commit after the DML execution.
    :param sync_session: execution option `synchronize_session`. Default False.

    IMPORTANT: decorated function must not be async, but after decoration it
    becomes async.

    More info in the noorm.sqlalchemy_async docstring.
    """

    if callable(func):
        the_func = func
    else:
        the_func = None

    def decorator_wrapper():
        def decorator(
            func: Callable[F_Spec, Executable],
        ) -> Callable[Concatenate[AsyncSession, F_Spec], Coroutine[Any, Any, None]]:
            class wrapper(WrapperBase):
                async def __call__(
                    self,
                    session: AsyncSession,
                    *args: F_Spec.args,
                    **kwargs: F_Spec.kwargs,
                ) -> None:
                    with MetricsCollector(self._func):
                        if (
                            sql_stmt := req_sql_n_params(
                                self._func, args, kwargs, sync_session
                            )
                        ) is not None:
                            await session.execute(sql_stmt)
                            await _commit_if_needed(session, sql_stmt, no_commit)

            return wrapper(func)

        return decorator

    wrap_decorator = decorator_wrapper()
    if callable(func):
        return wrap_decorator(the_func)
    else:
        return wrap_decorator



"""
NoORM (Not Only ORM) helpers for pymysql
"""

from typing import Type, Callable, Generator, ParamSpec, TypeVar, Concatenate, overload

from pymysql import Connection

from .._common import WrapperBase
from .._db_api_2 import PrepareFuncResult, req_sql_n_params
from ..registry import MetricsCollector

F_Spec = ParamSpec("F_Spec")
TR = TypeVar("TR")


def sql_fetch_all(row_type: Type[TR], sql: str | None = None):
    """
    Use this decorator to make a "fetch" SQL statement executor out of
    the function that prepares parameters for the query

    :param row_type: type of expected result. Usually some dataclass or named tuple
    :param sql: SQL statement to execute. If None, the SQL statement must be provided
    by decorated function.

    More info in the noorm.pymysql docstring.
    """

    def decorator(
        func: Callable[F_Spec, PrepareFuncResult | None]
    ) -> Callable[Concatenate[Connection, F_Spec], list[TR]]:
        class wrapper(WrapperBase):
            def __call__(
                self, conn: Connection, *args: F_Spec.args, **kwargs: F_Spec.kwargs
            ) -> list[TR]:
                with MetricsCollector(self._func) as mc:
                    if sql_and_params := req_sql_n_params(
                        self._func, args, kwargs, sql
                    ):
                        with conn.cursor() as cur:
                            cur.execute(*sql_and_params)
                            col_names = tuple(el[0] for el in cur.description)
                            res: list[TR] = [
                                row_type(**{n: v for n, v in zip(col_names, r)})
                                for r in cur
                            ]
                            mc.tuples = len(res)
                            return res
                    return []

        return wrapper(func)

    return decorator


def sql_iterate(row_type: Type[TR], sql: str | None = None):
    """
    Use this decorator to make a query and iterate through results. Be careful with
    this feature and, if possible, use `sql_fetch_all` instead, because
    `sql_fetch_all` gives you less possibilites to shoot your leg.

    :param row_type: type of expected result. Usually some dataclass or named tuple
    :param sql: SQL statement to execute. If None, the SQL statement must be provided
    by decorated function.

    More info in the noorm.pymysql docstring.
    """

    def decorator(
        func: Callable[F_Spec, PrepareFuncResult | None]
    ) -> Callable[Concatenate[Connection, F_Spec], Generator[TR, None, None]]:
        class wrapper(WrapperBase):
            def __call__(
                self, conn, *args: F_Spec.args, **kwargs: F_Spec.kwargs
            ) -> Generator[TR, None, None]:
                with MetricsCollector(self._func) as mc:
                    if sql_and_params := req_sql_n_params(
                        self._func, args, kwargs, sql
                    ):
                        with conn.cursor() as cur:
                            cur.execute(*sql_and_params)
                            col_names = tuple(el[0] for el in cur.description)
                            is_first_row = True
                            for r in cur:
                                if is_first_row:
                                    mc.finish(None)
                                    is_first_row = False
                                yield row_type(**{n: v for n, v in zip(col_names, r)})

        return wrapper(func)

    return decorator


def sql_one_or_none(row_type: Type[TR], sql: str | None = None):
    """
    Use this decorator to make an "fetchrow" SQL statement executor out of
    the function that prepares parameters for the query

    :param row_type: type of expected result. Usually some dataclass or named tuple
    :param sql: SQL statement to execute. If None, the SQL statement must be provided
    by decorated function.

    More info in the noorm.pymysql docstring.
    """

    def decorator(
        func: Callable[F_Spec, PrepareFuncResult | None],
    ) -> Callable[Concatenate[Connection, F_Spec], TR | None]:
        class wrapper(WrapperBase):
            def __call__(
                self, conn: Connection, *args: F_Spec.args, **kwargs: F_Spec.kwargs
            ) -> TR | None:
                with MetricsCollector(self._func) as mc:
                    if sql_and_params := req_sql_n_params(
                        self._func, args, kwargs, sql
                    ):
                        with conn.cursor() as cur:
                            cur.execute(*sql_and_params)
                            col_names = tuple(el[0] for el in cur.description)
                            for row in cur:
                                mc.tuples = 1
                                return row_type(
                                    **{n: v for n, v in zip(col_names, row)}
                                )
                    return None

        return wrapper(func)

    return decorator


def sql_scalar_or_none(res_type: Type[TR], sql: str | None = None):
    """
    Use this decorator to make a "scalar" SQL statement executor out of
    the function that prepares parameters for the query

    :param res_type: type of expected result. For scalar queries it is usually `int`,
    `str`, `bool`, `datetime`, or whatever can be produced by scalar query.
    :param sql: SQL statement to execute. If None, the SQL statement must be provided
    by decorated function.

    More info in the noorm.pymysql docstring.
    """

    def decorator(
        func: Callable[F_Spec, PrepareFuncResult | None],
    ) -> Callable[Concatenate[Connection, F_Spec], TR | None]:
        class wrapper(WrapperBase):
            def __call__(
                self, conn: Connection, *args: F_Spec.args, **kwargs: F_Spec.kwargs
            ) -> TR | None:
                with MetricsCollector(self._func) as mc:
                    if sql_and_params := req_sql_n_params(
                        self._func, args, kwargs, sql
                    ):
                        with conn.cursor() as cur:
                            cur.execute(*sql_and_params)
                            for row in cur:
                                mc.tuples = 1
                                return row[0]
                    return None

        return wrapper(func)

    return decorator


def sql_fetch_scalars(res_type: Type[TR], sql: str | None = None):
    """
    Use this decorator to make a "fetch scalars" SQL statement executor out of
    the function that prepares parameters for the query

    :param row_type: type of expected result. For scalar queries it is usually `int`,
    `str`, `bool`, `datetime`, or whatever can be produced in a single-column query
    result.
    :param sql: SQL statement to execute. If None, the SQL statement must be provided
    by decorated function.

    More info in the noorm.pymysql docstring.
    """

    def decorator(
        func: Callable[F_Spec, PrepareFuncResult | None]
    ) -> Callable[Concatenate[Connection, F_Spec], list[TR]]:
        class wrapper(WrapperBase):
            def __call__(
                self, conn: Connection, *args: F_Spec.args, **kwargs: F_Spec.kwargs
            ) -> list[TR]:
                with MetricsCollector(self._func) as mc:
                    if sql_and_params := req_sql_n_params(
                        self._func, args, kwargs, sql
                    ):
                        with conn.cursor() as cur:
                            cur.execute(*sql_and_params)
                            res: list[TR] = [r[0] for r in cur]
                            mc.tuples = len(res)
                            return res
                    return []

        return wrapper(func)

    return decorator


def sql_iterate_scalars(res_type: Type[TR], sql: str | None = None):
    """
    Use this decorator to make a query and iterate through scalar results. Be careful
    with this feature and, if possible, use `sql_fetch_scalars` instead, because
    `sql_fetch_scalars` gives you less possibilites to shoot your leg.

    :param res_type: type of expected result. For scalar queries it is usually `int`,
    `str`, `bool`, `datetime`, or whatever can be produced in a single-column query
    result.
    :param sql: SQL statement to execute. If None, the SQL statement must be provided
    by decorated function.

    More info in the noorm.pymysql docstring.
    """

    def decorator(
        func: Callable[F_Spec, PrepareFuncResult | None]
    ) -> Callable[Concatenate[Connection, F_Spec], Generator[TR, None, None]]:
        class wrapper(WrapperBase):
            def __call__(
                self, conn, *args: F_Spec.args, **kwargs: F_Spec.kwargs
            ) -> Generator[TR, None, None]:
                with MetricsCollector(self._func) as mc:
                    if sql_and_params := req_sql_n_params(
                        self._func, args, kwargs, sql
                    ):
                        with conn.cursor() as cur:
                            cur.execute(*sql_and_params)
                            is_first_row = True
                            for r in cur:
                                if is_first_row:
                                    mc.finish(None)
                                    is_first_row = False
                                yield r[0]

        return wrapper(func)

    return decorator


@overload
def sql_execute(
    func: Callable[F_Spec, PrepareFuncResult | None]
) -> Callable[Concatenate[Connection, F_Spec], None]:
    pass  # pragma: no cover


@overload
def sql_execute(
    sql: str | None = None,
) -> Callable[
    [Callable[F_Spec, None]], Callable[Concatenate[Connection, F_Spec], None]
]:
    pass  # pragma: no cover


def sql_execute(  # type: ignore
    sql: Callable[F_Spec, PrepareFuncResult | None] | str | None = None,
):
    """
    Use this decorator to execute a statement without responding a result

    :param sql: SQL statement to execute. If None, the SQL statement must be provided
    by decorated function.

    More info in the noorm.pymysql docstring.
    """

    if callable(sql):
        the_func = sql
        the_sql = None
    else:
        the_func = None
        the_sql = sql

    def decorator_wrapper(sql: str | None):
        def decorator(
            func: Callable[F_Spec, PrepareFuncResult | None],
        ) -> Callable[Concatenate[Connection, F_Spec], None]:
            class wrapper(WrapperBase):
                def __call__(
                    self, conn: Connection, *args: F_Spec.args, **kwargs: F_Spec.kwargs
                ) -> None:
                    with MetricsCollector(self._func):
                        if sql_and_params := req_sql_n_params(
                            self._func, args, kwargs, sql
                        ):
                            with conn.cursor() as cur:
                                cur.execute(*sql_and_params)

            return wrapper(func)

        return decorator

    wrap_decorator = decorator_wrapper(the_sql)
    if callable(sql):
        return wrap_decorator(the_func)
    else:
        return wrap_decorator



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['noorm',
 'noorm.aiomysql',
 'noorm.aiosqlite',
 'noorm.asyncpg',
 'noorm.psycopg2',
 'noorm.pymysql',
 'noorm.sqlalchemy_async',
 'noorm.sqlalchemy_sync',
 'noorm.sqlite3']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'true-noorm',
    'version': '0.1.5',
    'description': 'NoORM (Not only ORM) - make your database operations convenient and natural',
    'long_description': '[![PyPI pyversions](https://img.shields.io/pypi/pyversions/true-noorm.svg)](https://pypi.python.org/pypi/true-noorm/)\n[![PyPI status](https://img.shields.io/pypi/status/true-noorm.svg)](https://pypi.python.org/pypi/true-noorm/)\n[![GitHub license](https://img.shields.io/github/license/Naereen/StrapDown.js.svg)](https://github.com/amaslyaev/noorm/blob/master/LICENSE)\n[![codecov](https://codecov.io/gh/amaslyaev/noorm/graph/badge.svg?token=31YWXNHPMM)](https://codecov.io/gh/amaslyaev/noorm)\n[![Code style](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black/)\n\n## noorm\n\nNoORM (Not only ORM) - make your database operations convenient and natural\n\n## Install\n\n**noorm** requires Python 3.10 or newer. Install it from PyPI:\n```shell\n$ pip install true-noorm\n```\nPlease note that the correct name is "**true**-noorm".\n\n## NoORM principles\n\n1. It is not a holy war against ORM but "in addition to".\n2. It is not one more "finally perfect" ORM. It is not an ORM at all. **No persistent objects, no "ideal" entities.**\n3. It should be good for for medium-sized and big project.\n4. Focus on developer experience.\n5. Not only a set of helpers to write less code, but first of all, an approach that guides to more understandable, performant, scalable, robust, and maintainable solutions.\n\n## Usage\n\nImpotring `noorm` depends on DB you use in your project. Available options:\n\n- `import noorm.sqlite3 as nm` – for SQLite\n- `import noorm.aiosqlite as nm` – for asynchronous SQLite via **aiosqlite**\n- `import noorm.psycopg2 as nm` – for synchronous Postgres via **psycopg2**\n- `import noorm.asyncpg as nm` – for asynchronous Postgres via **asyncpg**\n- `import noorm.pymysql as nm` – for synchronous MySQL/MariaDB via **PyMySQL**\n- `import noorm.aiomysql as nm` – for asynchronous MySQL/MariaDB via **aiomysql**\n- `import noorm.sqlalchemy_sync as nm` – for synchronous **SqlAlchemy**\n- `import noorm.sqlalchemy_async as nm` – for asynchronous **SqlAlchemy**\n\nYes, using the SqlAlchemy ORM through NoORM is a nice idea.\n\nAfter importing "nm" you use `@nm.sql_...` decorators to create functions that perform database operations. All other your application code uses these functions as a so-called "DB API layer". The decorators are:\n- **@nm.sql_fetch_all** – to make a query and produce a list of objects of specified type. The query is usually SELECT, but it is also useful with data manipulations RETURNING data.\n- **@nm.sql_one_or_none** – to make a query and produce one object or None if nothing is found.\n- **@nm.sql_scalar_or_none** – to get a scalar or None if nothing is found.\n- **@nm.sql_fetch_scalars** – to get a list of scalars.\n- **@nm.sql_execute** – to execute something, usually INSERT, UPDATE, or DELETE.\n- **@nm.sql_iterate** and **@nm.sql_iterate_scalars** – to make a query and iterate through results – objects or scalars respectively. Be careful with this features and, if possible, use `sql_fetch_all` and `sql_fetch_scalars` instead, because they give you less possibilites to shoot your leg. These functions are not implemented for **asyncpg**.\n\nUsage of these decorators in different submodules and underlying databases might have own peculiarities, so check docstring documentation of the chosen "nm".\n\n#### Example for SQLite through the sqlite3 standard library\n\nAssume we have a **users** table with fields:\n- Integer `id` (`rowid` in SQLite)\n- String `username`\n- String `email`\n\nAnd an **orders** table:\n- Integer `id` (`rowid` in SQLite)\n- Integer `user_id` references to user id\n- Datetime `order_date` (TEXT in SQLite)\n- Decimal `amount` (TEXT in SQLite)\n\n```python\nfrom dataclasses import dataclass\nfrom decimal import Decimal\nfrom datetime import datetime\nimport sqlite3\nimport noorm.sqlite3 as nm\n\n@dataclass\nclass DbUser:  # When we need only basic info. Not a "model"! Just a dataclass!\n    id: int\n    username: str\n    email: str\n\n@nm.sql_fetch_all(DbUser, "SELECT rowid AS id, username, email FROM users")\ndef get_all_users():\n    pass  # no parameters, so just "pass"\n\n@nm.sql_one_or_none(\n    DbUser, "SELECT rowid AS id, username, email FROM users WHERE rowid = :id"\n)\ndef get_user_by_id(id_: int):\n    return nm.params(id=id_)\n\n@dataclass\nclass DbUserWithOrdersSummary:  # With additional info from `orders` table\n    id: int\n    username: str\n    sum_orders: Decimal | None  # SQLite noorm can make decimal out of TEXT\n    first_order: datetime | None  # and datatime too.\n    last_order: datetime | None\n\n@nm.sql_fetch_all(\n    DbUserWithOrdersSummary,\n    """SELECT\n        u.rowid AS id, u.username,\n        SUM(o.amount) AS sum_orders,\n        MIN(o.order_date) AS first_order, MAX(o.order_date) AS last_order\n    FROM users u\n        LEFT OUTER JOIN orders o ON o.user_id = u.rowid\n    GROUP BY u.rowid, u.username ORDER BY u.rowid\n    """,\n)\ndef get_users_with_order_summary():\n    pass\n\ndef main():\n    with sqlite3.connect("data.sqlite") as conn:\n        # will use our DB API functions\n        for usr in get_all_users(conn):\n            print(usr)\n\n        print(f"{get_user_by_id(conn, 1)=}")\n\n        for usr_summary in get_users_with_order_summary(conn):\n            print(usr_summary)\n```\n\n#### Example for SQLite through SqlAlchemy\n\nWill use asynchronous version of SqlAlchemy "nm".\n```python\nimport sqlalchemy as sa\nfrom sqlalchemy.ext.asyncio import create_async_engine, AsyncSession\nimport noorm.sqlalchemy_async as nm\n...\n\n@nm.sql_fetch_all(DbUser)\ndef get_all_users():  # Notice no "async"\n    return sa.select(User.id, User.username, User.email)\n\n@nm.sql_one_or_none(DbUser)\ndef get_user_by_id(id_: int):\n    return sa.select(User.id, User.username, User.email).where(User.id == id_)\n\n@nm.sql_fetch_all(DbUserWithOrdersSummary)\ndef get_users_with_order_summary():\n    return (\n        sa.select(\n            User.id,\n            User.username,\n            sa.func.sum(Order.amount).label("sum_orders"),\n            sa.func.min(Order.order_date).label("first_order"),\n            sa.func.max(Order.order_date).label("last_order"),\n        )\n        .select_from(User)  # `User` and `Order` are ORM model classes\n        .outerjoin(Order, Order.user_id == User.id)\n        .group_by(User.id)\n        .order_by(User.id)\n    )\n\nasync def main():\n    engine = create_async_engine("sqlite+aiosqlite:///data.sqlite")\n    async with AsyncSession(engine) as session:\n        for usr in await get_all_users(session):  # Notice "await"\n            print(usr)\n\n        print(f"user_1={await get_user_by_id(session, 1)}")\n\n        for usr_summary in await get_users_with_order_summary(session):\n            print(usr_summary)\n```\n\n## Suggestions\n\n1. Code structure:\n   - Avoid scattering DB API functions throughout the codebase. It\'s preferable to consolidate them in dedicated locations.\n   - For small applications, consider housing all DB API functions in a single module, which becomes the DB API layer.\n   - In larger applications, dividing the DB API layer into multiple modules is advisable. For example, organize user management functions in `db_api/users.py` and order processing functions in `db_api/orders.py`.\n   - For systems with distinct, independent subsystems, consider placing common functions in a shared location, such as the `db/db_api/` folder, and specific functions in subsystem folders like `external_api/db_api/`.\n   - Declare the "Db..." dataclasses immediately preceding the functions that produce them.\n2. Naming:\n   - Prefix classes returned from DB API functions with "Db". For instance, use "DbUsers", "DbOrders", "DbOrdersWithDetails", "DbInvoicesReportLine".\n   - Prefix functions that SELECT data from the DB with "get_". For example, "get_user_by_id", "get_orders_by_user", etc.\n   - Use prefix "iter_" for DB API functions made using `@nm.sql_iterate` and `@nm.sql_iterate_scalars` decorators.\n   - For data manipulation functions:\n     - Use prefixes "ins_", "upd_", "del_" for INSERTs, UPDATEs, DELETEs respectively.\n     - Employ "upsert_" for upsert operations.\n\n## Advanced features\n\n#### Cancelling operations\n\nIf, for any reason, you need to terminate execution and produce a default result in your function, raise the `nm.CancelExecException` exception. Example:\n```python\n@nm.sql_fetch_all(DbOrder)\ndef get_orders_by_ids(order_ids: list[int]):\n    if not order_ids:\n        raise nm.CancelExecException\n    return select(Order.id, Order.date, Order.amount).where(\n        Order.id.in_(order_ids)  # <<< empty list is not acceptable here\n    )\n```\nThe `nm.CancelExecException` triggers production of a default empty result without querying the DB:\n- `@nm.sql_fetch_all` and `@nm.sql_fetch_scalars` return an empty list\n- `@nm.sql_one_or_none` and `@nm.sql_scalar_or_none` return None\n- `@nm.sql_execute` takes no action\n\n#### Registry\n\nObservability is crucial. This library facilitates collecting statistics on DB API function usage out of the box. Statistics include:\n- **calls** – number of calls\n- **duration** – total execution time\n- **tuples** – total number of retrieved tuples\n- **fails** – number of failed calls\n- **fails_by_error** – dict[str, int] – detailed fails categorized by exception types\n```python\nfrom noorm.registry import get_registry\nregistry = get_registry()\nstat = registry.stat_by_name["db.db_api.orders.get_orders_by_user"]\nprint(stat)  # Stat(calls=3, duration=0.0324, tuples=11, fails=0, fails_by_error={})\n```\nTo collect statistics in a multiprocessing application, initialize this option in your MainProcess:\n```python\n# Example for uvicorn\nimport uvicorn\nfrom noorm.registry import get_registry\n\nasync def app(scope, receive, send):\n    ...\n\nif __name__ == "__main__":\n    get_registry().init_multiprocess_registry()  # <<< here\n    uvicorn.run("main:app", port=5000, workers=3, log_level="info")\n```\nConsequently, all DB operations occurring in child processes will be aggregated in the MainProcess registry.\n\n**Important**: statistics for `@nm.sql_iterate` and `@nm.sql_iterate_scalars` is not precise:\n1. `stat.duration` is counted only for query execution and first row extraction.\n2. `stat.tuples` is always zero.\n3. `stat.fails` and `stat.fails_by_error` do not counter errors that might happen after successful first row extraction.\n\n#### Executing unwrapped functions\n\nTo call an unwrapped version of a DB API function for evaluation, testing, or debugging purposes:\n```python\norders_list = await get_orders_by_user(session, 1)  # A "normal" call\norders_list_q = get_orders_by_user.unwrapped(1)  # An "unwrapped" call\nprint(str(orders_list_q.compile()))  # Want to see a compiled SqlAlchemy "SELECT ..."\n```\n\n## Acknowledgements\n\nInspired and sponsored by [FRAMEN](https://www.framen.com/)\n',
    'author': 'Alexander Maslyeav',
    'author_email': 'maslyaev@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/amaslyaev/noorm/',
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)
